#!python
import ABIFReader, DNA, Fasta
import re, sys, os

read_failed = 'FAILED'
read_lqv = 'LOW_QUALITY'
read_os = 'OVER_SHOT'
read_short = 'SHORT'

class ABISeq:
	def __init__(self, abifile, qscore=30, force=1, seqlen=300):
		if not os.path.isfile(abifile):
			print "ERROR: file '%s' does not exist !!" %  abifile
			sys.exit()

		try:
			reader = ABIFReader.ABIFReader(abifile)
		except:
			print "ERROR: invalid ABI format for file %s" % abifile
			sys.exit()

		# this sample is the name we import into sequencer
		# for some of our file, it could be 010208A_2008-01-02_A02_016.ab1
		plateid, well, sample, bases, quals, signal2noice = reader.getSeqInfo() 

		reader.close()

		# filename
		self.file = abifile
		self.filename = re.sub('.ab1', '', os.path.basename(self.file) )

		s2n = sum(signal2noice)/len(signal2noice)
		self.s2n = s2n
		if s2n < 40:
			self.status = read_failed
		elif s2n > 3000:
			self.status = read_os
		else:
			s_ = int(sum(quals)/len(quals))
			if s_ <= qscore:
				self.status = read_lqv
			else:
				self.status = None


		# plateid, well, and sample
		self.plateid = plateid
		r = well[0]
		c = eval(well.replace(r, ''))
		self.well = '%s%02d' % (r,c)

		# force sample to be the same as file name
		basename = os.path.basename(abifile).replace('.ab1', '')
		m = re.search('_\d{4}\-\d\d\-\d\d', basename)

		# force sample name to be the same as file name
		if force:
			if m:
				sample = basename[:m.start()]
			else:
				sample = basename

			while sample.endswith('_'):
				sample = sample[:-1]
			self.sample = sample.strip()
		else:
			self.sample = sample

		# bases and quality score
		self.bases = bases
		self.quals = quals

		if len(self.bases) != len(self.quals):
			print 'ERROR: parsing abi file: %s, count(bases) != count(quality_values)' % abifile
			sys.exit()

		self.trimstart = None
		self.trimend = None
		self.trimscore = None

		"""
		reformat sample for cap3 application
		"""
		template= primer =''
		cols = self.sample.split('_')
		primer = cols[-1]
		template = self.sample.replace('_%s'% primer, '')
		self.template = template
		self.primer = primer
		self.cap3_name = self.sample
		#"%s.%s"% (template, primer)
		#print "SAMPLE NAME:%s, cap3 %s" % ( self.sample, self.cap3_name)

	def __str__(self):
		tmp = ['file:%s, sample:%s, well:%s, trimscore=%s, readlen=%s' % (self.file, self.sample, self.well, self.score(), self.trimlength())] 
		return str(tmp)

	def getsample(self):
		# specific for EMD report
		s = self.sample
		# remove primer information
		pat = '[A-H]\d\d[HL]'
		m = re.search(pat, s)
		if not m:
			return s
		end = m.end()
		return s[:end]


	def failed(self):
		if self.status in [read_lqv, read_failed]:
			return 1
		return 0

	def trim_by_position(self, start=20, end=800):
		self.trimstart = start
		self.trimend = min(len(self.bases), end)
		quals = self.quals[self.trimstart:self.trimend]
		self.trimscore =  int(0.5+sum(quals)*1.0/len(quals))

		if self.trimscore < 30:
			self.status = read_lqv
		elif self.trimend - self.trimstart + 1 < seqlen:
			self.status = read_short


	def trim_by_quality(self, quality_cutoff=20, base_count=4):
		"""
		default trim function, trim using quality values
		remove bases from both ends until fewer than 3/20 have QV < 20
		"""
		window = 20
		quality_cutoff = 20
		base_count = 4

		if self.status and self.status == read_failed:
			return

		# trim 5'end 
		Ncnt5=Ncnt3=0
		for i in range(0, len(self.quals)):
			tmplist = self.quals[i:i+window]
			mycnt = 0
			for x in tmplist:
				if x<quality_cutoff:
					mycnt += 1
			bases = self.bases[i:i+window]
			bases = bases.upper()
			bases = re.sub('[AGCT]', '', bases)
			Ncnt5 = len(bases)
			if mycnt < base_count and Ncnt5 < base_count:
				break

		for start in range(i, len(self.quals)):
			if self.quals[start] >= quality_cutoff:
				break
		# trim 3'end
		for i in range(1, len(self.quals)):
			tmplist = self.quals[-(i+window):-i]
			mycnt = 0
			for x in tmplist:
				if x<quality_cutoff:
					mycnt += 1
				
			bases = self.bases[-(i+window):-i]
			bases = bases.upper()
			Ncnt3 = len(bases)
			bases = re.sub('[AGCT]', '', bases)
			Ncnt3 = len(bases)
			if mycnt < base_count and Ncnt3 < base_count:
				break
		j = i	
		for i in range(j, len(self.quals)):
			t = self.quals[-i]
			if t >= quality_cutoff:
				break
		end = len(self.quals)-i
		if end <= start:
			self.status = read_failed
			return 
		self.trimstart = start
		self.trimend = end
		quals = self.quals[self.trimstart:self.trimend]
		self.trimscore =  int(0.5+sum(quals)*1.0/len(quals))

		if self.trimscore < 30:
			self.status = read_lqv
		elif self.trimend - self.trimstart + 1 < 600:
			self.status = read_short

	def trimseq(self, format=0):
		if not self.trimscore:
			return ""
		bases = self.bases[self.trimstart:self.trimend]
		if not format:
			return "".join(bases)
		tmp = []
		for i in range(0, len(bases), 60):
			tmp.append("".join(bases[i:i+60]) )
		return "\n".join(tmp)

	def trimlength(self):
		if not self.trimstart or not self.trimend:
			self.trim_by_quality()
		if not self.trimstart or not self.trimend:
			return 0

		x = self.trimend-self.trimstart+1
		#print "TRIM LENGTH:", x
		return x

	def LOR(self):
		if not self.trimstart or not self.trimend:
			return 0
		return self.trimend-self.trimstart+1

	def Q20(self, score=20):
		scores = self.quals[self.trimstart:self.trimend]
		cnt = 0
		for s in scores:
			if s > score:
				cnt +=1
		return cnt
		
	def trimqual(self, format=0):
		if not self.trimstart or not self.trimend:
			self.trim_by_quality()
		scores = self.quals[self.trimstart:self.trimend]
		scores = [str(x) for x in scores]
		if not format:
			return " ".join(scores)

		tmp = []
		for i in range(0, len(scores), 60):
			tmp.append(" ".join(scores[i:i+60]) )
		return "\n".join(tmp)


	def score(self):
		if self.trimstart and self.trimend:
			quals = self.quals[self.trimstart:self.trimend]
		else:
			quals = self.quals
		return int(sum(quals)/len(quals))

	def fasta(self, trim=0, rc=0):
		return Fasta.Fasta(self.sample, self.seq(trim,rc))

	def seq(self, trim=0, rc=0):
		if trim:
			if not self.trimstart:
				self.trim_by_quality()
			if not self.trimstart:
				return ('','')
			
			bases = self.bases[self.trimstart:self.trimend]
		else:
			bases = self.bases

		seq = "".join(bases).upper()
		return seq

	def qual(self, trim=0, rc=0):
		quals = self.quals
		if trim:
			if not self.trimstart:
				self.trim_by_quality()
			if not self.trimstart:
				return ('', '')
			quals = self.quals[self.trimstart:self.trimend]

		if rc:
			quals.reverse()

		quals = [str(x) for x in quals]
		return " ".join(quals)

	def _template_primer(self, i, sep='_'):
		"""
		default name format
		"""
		t = self.sample[:i]
		j = t.rfind(sep)
		template = t[:j]
		primer = t[(i+1):]
		return (template, primer)
		
	def get_template_primer(self, sep='_'):
		datepat = "_\d{4}\-\d{2}\-\d{2}"
		m = re.match(datepat, self.sample)
		if m:
			return self._template_primer(m.start(), sep)

		#print 'cap3_name:', self.cap3_name, self.sample
		#cols = self.cap3_name.split('.')
		cols = self.sample.split(sep)
		#template = cols[0]
		#primer = self.sample.replace('%s_' % template, '')
		primer = cols[-1]
		template = self.cap3_name.replace('.%s' % primer, '')
		template = self.sample.replace('%s%s' % (sep, primer), '')
		return (template, primer)

	def match_primer(self, primer):
		t, pr = self.get_template_primer()
		if pr.lower() == primer.lower():
			return 1
		return 0

def test(abifile):
	myseq = ABISeq(abifile)
	print "cap3 name:", myseq.cap3_name, myseq.well
	print "trimseq:", myseq.trimseq()
	print "trimscore:", myseq.trimscore

	myseq.trim_by_quality()
	print "trimseq:", myseq.trimseq(), len(myseq.trimseq())
	print "trimscore:", myseq.trimscore, myseq.trimstart, myseq.trimend

if __name__ == '__main__':
	test(sys.argv[1])
