#!python

import DNA, Object

import os, sys, re, glob, getopt
import Fasta

blast_fields_m8 = [ 'query', 'subj', 'ident', 'length', 'miss', 'gap', 'qstart', 'qend', 'sstart', 'send', 'evalue', 'score']
NUM_FIELDS = ['ident', 'length', 'miss', 'gap', 'qstart', 'qend', 'sstart', 'send', 'evalue', 'score']

class BlastHSP(Object.Object):
	# Fields: Query id, Subject id, % identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score

	# these are the fields in blast result

	def __init__(self, line):
		cols = line.split('\t')
		d = dict(zip(blast_fields_m8, cols))

		Object.Object.__init__(self, **d)

		for k, v in d.items():
			if k in NUM_FIELDS:
				v = eval(v)
			self.__dict__[k] = v

	def set(self, line):
		cols = [x.strip() for x in line.split('\t')]
		if len(cols) < 12:
			return
		self.__dict__.update(dict(zip(blast_fields_m8, cols)))
		for k in NUM_FIELDS:
			self.__dict__[k] = eval(self.__dict__[k])

	def format(self):
		tmp= [str(self.__dict__[k]) for k in blast_fields_m8]
		return "\t".join(tmp)


	def __str__(self):
		return Object.Object.__str__(self)	

class BlastHit(Object.Object):
	"""
	default param: name, hsps
	"""
	def __init__(self, **kw):
		Object.Object.__init__(self, **kw)	

	def __str__(self):
		tmp = [str(x) for x in self.hsps]
		return "\n".join(tmp)

	def __getattr__(self, k):
		if self.__dict__.has_key(k):
			return self.__dict__[k] 
		self.hsps.sort(lambda x,y: cmp(y.score, x.score))
		return getattr(self.hsps[0], k)

	def get_best_HSPs(self):
		self.hsps.sort(lambda x,y: cmp(y.score, x.score))
		best = []
		best.append(self.hsps[0])
		for h in self.hsps[1:]:
			if h.score + 15 >= best[0].score:
				best.append(h)
			else:
				break
		return best


class BlastResult(Object.Object):
	def __init__(self, **kw):
		Object.Object.__init__(self, **kw)

	def __str__(self):
		tmp = ['%s(%s)' % (self.__class__.__name__, self.query)]
		for h in self.hits:
			tmp.append(str(h))
		return "\n".join(tmp)

	def get_best_hits(self):
		best_hits = []

		for h in self.hits:
			h.hsps.sort(lambda x,y: cmp(y.score, x.score))	
			h.score = h.hsps[0].score
		self.hits.sort(lambda x,y: cmp(y.score, x.score))

		best = self.hits[0]
		best_hits.append(best)

		for h in self.hits[1:]:
			if h.score + 15 >= best.score:
				best_hits.append(h)
			else:
				break

		return best_hits
	
def parse(blastoutfile, blast_output_format='-m 8'):
	if not os.path.isfile(blastoutfile):
		print "ERROR: there is no blast result file, that usually means either the reference sequence is wrong, or the sequencing data does not match to the reference."
		sys.exit()
	if blast_output_format in ['-m 8', '-m 9']:
		return parse_m8and9(blastoutfile)

def parse_m8and9(blastoutfile, headers=blast_fields_m8, identity_cutoff=100):
	lines = open(blastoutfile).readlines()

	queryd = {}
	queries = []
	for line in lines:
		line = line.strip()
		if re.match('#', line):
			continue

		hsp = BlastHSP(line)
		
		queryd.setdefault(hsp.query, {'subjs':[]}) 
		if not hsp.subj in queryd[hsp.query]['subjs']:
			queryd[hsp.query]['subjs'].append(hsp.subj)
		queryd[hsp.query].setdefault(hsp.subj, []).append(hsp)
		if not hsp.query in queries:
			queries.append(hsp.query)

	all_results = []
	for query in queries:
		subjd = queryd[query]
		subjs = subjd['subjs']
		hits = []
		for subj in subjs:
			hits.append( BlastHit(name = subj, hsps = subjd[subj]) )

		all_results.append( BlastResult(query = query, hits = hits) )

	return all_results

def get_hit_seqs(blastres, infile, dbfile, full_length=0):
	#print "getting fasta from %s" % infile
	queryseqs = Fasta.all(infile)

	queryd = {}
	for x in queryseqs:
		#print type(x), x
		n = x.name
		s = x.sequence
		if not n:
			continue
		queryd[n] = s

	dbd = {}
	alls = Fasta.all(dbfile)
	for a in alls:
		n = a.name
		s = a.sequence
		if not n:
			continue
		dbd[n] = s

	#print "\nParsing blast result file:", blastres
	res = parse(blastres)

	subjseq = []
	for r in res:
		query_seq = queryd[r.query]
		query_len = len(query_seq)
		tag = query_seq[:15].upper()
		etag = query_seq[-15:].upper()

		for h in r.hits:
			# how many hsps do we need to check for each hit? all
			# sort hsps by start coordinates
			hsps = h.hsps
			if not hsps:
				continue

			hsp = hsps[0]
			if full_length:
				if len(queryd[hsp.query]) - hsp.length >= 200:
					continue

			qstart = min([x.qstart for x in hsps])
			qend   = max([x.qend for x in hsps])
			rc = 0
			if hsp.sstart > hsp.send:
				rc = 1
			sbj_seq = dbd[hsp.subj]
			if len(hsps) == 1:
				hsp = h.hsps[0]
				sstart = min([hsp.sstart, hsp.send])
				send = max([hsp.sstart, hsp.send])
				#print  sstart, send
			else:
				all_coords = []
				for _h in hsps:
					if (rc and _h.sstart > _h.send) or (not rc and _h.sstart < _h.send):
						all_coords.extend([_h.sstart, _h.send])
				#print "2:",  all_coords
				sstart = min(all_coords)
				send = max(all_coords)
				#print sstart, send
			#print "BEFORE:", sstart, send, rc

			# adjust for first base
			if hsp.sstart < hsp.send:
				start = sstart + 1 - qstart
				#end =   send + query_len - qend
				end = start+query_len -1

				if start <= 0:
					start = 1
				myseq = sbj_seq[start-1:end]
			else:
				start = sstart + qend - query_len
				end = send + qstart -1
				#end = start+query_len +1
				if start <=0:
					start = 1
				myseq = sbj_seq[start-1:end]
				#print sbj_seq
				#print myseq
				myseq = DNA.DNA(myseq).reverse_complement()
				#print "b:", myseq
			#print "\nget_hit_seq:", hsp.query, "S:", hsp.subj, "start=%s, end=%s" % (sstart, send), len(myseq)
			subjseq.append(Fasta.Fasta(hsp.subj, myseq))

	return subjseq

def test(args):

	opts, ars = getopt.getopt(args, 'm:')
	#print opts, ars
	res = parse(ars[0])
	for r in res:
		print r.query
		for h in r.hits:
			hsp = h.hsps[0]
			print h.subj, hsp.qstart, hsp.qend, hsp.sstart, hsp.send

if __name__ == '__main__':
	if len(sys.argv) < 1:
		print "Usage: %s [-m 8] blastout  " % sys.argv[0]
		sys.exit()

	#test(sys.argv[1:])
	r = sys.argv[1]; i = sys.argv[2]; d = sys.argv[3]

	get_hit_seqs(r, i, d)

			
		
